/*
 * Copyright (c) 2005-2007 Ondrej Palkovsky, Michal Kebrt
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup arm32
 * @{
 */
/** @file
 */

#include <interrupt.h>
#include <ipc/irq.h>
#include <console/chardev.h>
#include <arch/drivers/gxemul.h>
#include <console/console.h>
#include <sysinfo/sysinfo.h>
#include <print.h>

/** Address of devices. */
#define GXEMUL_VIDEORAM		0x10000000
#define GXEMUL_KBD_ADDRESS	0x10000000
#define GXEMUL_KBD_IRQ 		0

static chardev_t console;
static irq_t gxemul_irq;

static void gxemul_write(chardev_t *dev, const char ch);
static void gxemul_enable(chardev_t *dev);
static void gxemul_disable(chardev_t *dev);
static char gxemul_do_read(chardev_t *dev);

static chardev_operations_t gxemul_ops = {
	.resume = gxemul_enable,
	.suspend = gxemul_disable,
	.write = gxemul_write,
	.read = gxemul_do_read,
};

/** Putchar that works with gxemul */
void gxemul_write(chardev_t *dev, const char ch)
{
	*((char *) GXEMUL_VIDEORAM) = ch;
}

/* Called from getc(). */
void gxemul_enable(chardev_t *dev)
{
//	cp0_unmask_int(GXEMUL_KBD_IRQ);
}

/* Called from getc(). */
void gxemul_disable(chardev_t *dev)
{
//	cp0_mask_int(GXEMUL_KBD_IRQ);
}

/** Read character using polling, assume interrupts disabled */
static char gxemul_do_read(chardev_t *dev)
{
	char ch;

	while (1) {
		ch = *((volatile char *) GXEMUL_KBD_ADDRESS);
		if (ch) {
			if (ch == '\r')
				return '\n';
			if (ch == 0x7f)
				return '\b';
			return ch;
		}
	}
}

/** Process keyboard interrupt. */
static void gxemul_irq_handler(irq_t *irq, void *arg, ...)
{
	if ((irq->notif_cfg.notify) && (irq->notif_cfg.answerbox))
		ipc_irq_send_notif(irq);
	else {
		char ch = 0;
		
			ch = *((char *) GXEMUL_KBD_ADDRESS);
			if (ch =='\r')
				ch = '\n';
			if (ch == 0x7f)
				ch = '\b';
			chardev_push_character(&console, ch);
	}
}

static irq_ownership_t gxemul_claim(void)
{
	return IRQ_ACCEPT;
}

void gxemul_kbd_grab(void)
{
	ipl_t ipl = interrupts_disable();
	spinlock_lock(&msim_irq.lock);
	gxemul_irq.notif_cfg.notify = false;
	spinlock_unlock(&gxemul_irq.lock);
	interrupts_restore(ipl);
}

void gxemul_kbd_release(void)
{
	ipl_t ipl = interrupts_disable();
	spinlock_lock(&gxemul_irq.lock);
	if (gxemul_irq.notif_cfg.answerbox)
		gxemul_irq.notif_cfg.notify = true;
	spinlock_unlock(&gxemul_irq.lock);
	interrupts_restore(ipl);
}


/* Return console object representing msim console */
void gxemul_console(devno_t devno)
{
	chardev_initialize("msim_console", &console, &gxemul_ops);
	stdin = &console;
	stdout = &console;
	
	irq_initialize(&gxemul_irq);
	gxemul_irq.devno = devno;
	gxemul_irq.inr = GXEMUL_KBD_IRQ;
	gxemul_irq.claim = gxemul_claim;
	gxemul_irq.handler = gxemul_irq_handler;
	irq_register(&gxemul_irq);
	
//	cp0_unmask_int(GXEMUL_KBD_IRQ);
	
	sysinfo_set_item_val("kbd", NULL, true);
	sysinfo_set_item_val("kbd.devno", NULL, devno);
	sysinfo_set_item_val("kbd.inr", NULL, GXEMUL_KBD_IRQ);
	sysinfo_set_item_val("kbd.address.virtual", NULL, GXEMUL_KBD_ADDRESS);
}

/** @}
 */
